# $Id: cinput.rb,v 1.24 2002/02/03 14:30:48 toki Exp $
# --
# :
#   Curses ȤäʸϤ롣
#   ꤷʸդ롣
#   ƤӽФ Curses::cbreak, Curses::noecho ƤǤȡ
#   1 ХʸѡޥХʸΤȤϤޤäθƤʤ
#   ѹǤ롣
#
# 饹⥸塼:
#   class Curses::Input
#
# 饹⥸塼δط:
#   [Curses::Input]
#     <include>
#       Curses
#       Keymap (callback.rb)
#
# ƥ饹⥸塼:
#   [Curses::Input]
#     ERASE
#     CLEAR
#     CANCEL
#     ENTER
#     KEYMAP_TABLE
#
# ƥ饹⥸塼Υ᥽å:
#   [Curses::Input]
#     <class method>
#       new(win, chars, no_keymap=false)
#     <method>
#       x
#       x=(x)
#       y
#       y=(y)
#       input
#       input=(input)
#       plain_attr
#       plain_attr=(attr)
#       text_attr
#       text_attr=(attr)
#       erase
#       clear
#       cancel
#       enter
#       gets
#     <private method>
#       set_keymap_table
#
# Ȥ:
#   require "cinput"
#   require "curses"
#
#   Curses::init_screen
#   begin
#     Curses::cbreak
#     Curses::noecho
#     if (defined? Curses::stdscr.keypad) then
#       # Curses patch ([ruby-list:16319]) ȤäƤȤ
#       # Window  keypad  true ꤹ뤳ȡ
#       Curses::stdscr.keypad(true)
#     end
#     Curses::addstr("[END: quit or exit]\n")
#     input = nil
#     loop do
#       Curses::setpos(Curses::stdscr.cury, 0)
#       Curses::addstr("Input [A-Za-z]: ")
#       cin = Curses::Input::new(Curses::stdscr,
#			         (?A..?Z).to_a + (?a..?z).to_a)
#       cin.input = input || ""
#       input = cin.gets
#       break if (input == "quit" || input == "exit")
#       Curses::addstr("\n" + (input || "") + "\n")
#       Curses::refresh
#     end
#   ensure
#     Curses::close_screen
#   end
#

require "curses"
require "thread"
require "rmp3/callback"

module Curses
  class Input
    include Curses
    include Keymap

    def initialize(win, chars, no_keymap=false)
      @m = Mutex::new
      @win = win
      @x, @y = @win.curx, @win.cury
      @chars = chars
      @input = ""
      @keymap = {}
      @plain_attr = nil
      @text_attr = nil
      set_keymap_table unless no_keymap
    end

    attr_accessor :x
    attr_accessor :y
    attr_accessor :input
    attr_accessor :plain_attr
    attr_accessor :text_attr

    def erase
      @m.synchronize do
	@input.chop!
      end
    end

    def clear
      @m.synchronize do
	@input = ""
      end
    end

    def cancel
      throw(:CURSES_INPUT_GETS, nil)
    end

    def enter
      @m.synchronize do
	unless (@input.empty?) then
	  throw(:CURSES_INPUT_GETS, @input)
	else
	  cancel
	end
      end
    end

    def gets
      @win.setpos(@y, @x)
      @win.addstr(@input)
      max_len = (@win.curx - @x)
      catch(:CURSES_INPUT_GETS) do
	begin
	  loop do
	    @m.synchronize do
	      @win.setpos(@y, @x)
	      if (defined? @win.attrset) then
		# curses-color.patch [ruby-list:18163]
		if (@plain_attr) then
		  @win.attrset(@plain_attr)
		elsif (@text_attr) then
		  @win.attrset(A_NORMAL)
		end
	      end
	      @win.addstr(" " * max_len)
	      @win.setpos(@y, @x)
	      if (defined? @win.attrset) then
		# curses-color.patch [ruby-list:18163]
		if (@text_attr) then
		  @win.attrset(@text_attr)
		end
	      end
	      @win.addstr(@input)
	    end
	    @win.refresh
	    key = @win.getch
	    if (@chars.include? key) then
	      @input << key.chr
	    else
	      hook_key(key, self)
	    end
	    max_len = [ max_len, @win.curx - @x ].max
	  end
	ensure
	  if (defined? @win.attrset) then
	    # curses-color.patch [ruby-list:18163]
	    if (@plain_attr || @text_attr) then
	      @win.attrset(A_NORMAL)
	    end
	  end
	end
      end
    end

    # [NOTE]
    # Keymap module method (callback.rb)
    # -
    # 1. set_keymap(key, callback)
    # 2. set_keymap(key) { |cin| ... }
    # -
    # callback ϼΥ᥽åɤʤФʤʤ
    #   callback.call(input)
    #

    ERASE  = proc{ |cin| cin.erase }
    CLEAR  = proc{ |cin| cin.clear }
    CANCEL = proc{ |cin| cin.cancel }
    ENTER  = proc{ |cin| cin.enter }

    if (defined? Key) then
      # for curses.patch [ruby-list:16319]
      KEYMAP_TABLE = [
	# <key list>                          <callback>
	[ [ KEY_BACKSPACE, KEY_DC, ?\C-h ], ERASE ],
	[ [ ?\C-u ],                        CLEAR ],
	[ [ ?\C-g, ?\e ],                   CANCEL],
	[ [ ?\n ],                          ENTER ]
      ]
    else
      KEYMAP_TABLE = [
	# <key list> <callback>
	[ [ ?\C-h ], ERASE ],
	[ [ ?\C-u ], CLEAR ],
	[ [ ?\C-g ], CANCEL ],
	[ [ ?\n ],   ENTER ]
      ]
    end

    def set_keymap_table
      for keys, callback in KEYMAP_TABLE
	for key in keys
	  set_keymap(key, callback)
	end
      end
    end
    private :set_keymap_table
  end
end
