# $Id: cmenu.rb,v 1.28 2002/02/03 14:30:50 toki Exp $
# --
# :
#   Curses Ȥäƥ˥塼ɽƹܤ򤹤롣
#   ˥塼ɽ Curses::cbreak, Curses::noecho ƤǤȡ
#   ˥塼ιܤɽ⹹ǤΤǡåɤȤäƥ˥塼
#   ܤƱѹ뤳ȤǤ롣
#   ѹǤ롣
#
# 饹⥸塼:
#   class Curses::Menu
#
# 饹⥸塼δط:
#   [Curses::Menu]
#     <include>
#       Curses
#       Keymap (callback.rb)
#
# ƥ饹⥸塼:
#   [Curses::Menu]
#     NEXT_ITEM
#     PREV_ITEM
#     NEXT_PAGE
#     PREV_PAGE
#     ENTER
#     CANCEL
#     REDRAW
#     KEYMAP_TABLE
#
# ƥ饹⥸塼Υ᥽å:
#   [Curses::Menu]
#     <class method>
#       new(win, no_keymap=false)
#     <method>
#       plain_attr
#       plain_attr=(attr)
#       cursor_attr
#       cursor_attr=(attr)
#       list
#       list=(list)
#       index
#       index=(index)
#       refresh
#       redraw
#       next_item
#       prev_item
#       next_page
#       prev_page
#       enter
#       cancel
#       get
#     <private method>
#       draw_list
#       set_keymap_table
#
# Ȥ:
#   require "cmenu"
#   require "curses"
#   require "thread"
#
#   Curses::init_screen
#   begin
#     Curses::cbreak
#     Curses::noecho
#     if (defined? Curses::stdscr.keypad) then
#       # Curses patch ([ruby-list:16319]) ȤäƤȤ
#       # Window  keypad  true ꤹ뤳ȡ
#       Curses::stdscr.keypad(true)
#     end
#     menu = Curses::Menu::new(Curses::stdscr)
#     Thread::new do
#       100.times do |i|
#         sleep 0.1
#         menu.list << i.to_s
#         menu.refresh
#       end
#     end
#     index = menu.get
#   ensure
#     Curses::close_screen
#   end
#   p index
#

require "curses"
require "thread"
require "rmp3/callback"

module Curses
  class Menu
    include Curses
    include Keymap

    def initialize(win, no_keymap=false)
      @m = Mutex::new
      @win = win
      @list = []
      @index = 0
      @keymap = {}
      @plain_attr = nil
      @cursor_attr = nil
      set_keymap_table unless no_keymap
    end

    attr_accessor :plain_attr
    attr_accessor :cursor_attr

    def list
      @m.synchronize do
	@list
      end
    end

    def list=(list)
      @m.synchronize do
	@list = list
      end
    end

    def index
      @m.synchronize do
	@index
      end
    end

    def index=(index)
      @m.synchronize do
	@index = index
      end
    end

    def get_drawn_range
      nitems = @win.maxy
      pos = @index / nitems
      min = pos * nitems
      max = [ (pos + 1) * nitems - 1, @list.length - 1 ].min
      [ min, max ]
    end
    private :get_drawn_range

    def draw_list
      @m.synchronize do
	if (@list.empty?) then
	  @index = 0
	else
	  if (@index > @list.length) then
	    @index = @list.length - 1
	  end
	end
	if (defined? @win.attrset) then
	  # for curses-color.patch [ruby-list:18163]
	  if (@plain_attr) then
	    @win.attrset(@plain_attr)
	  end
	end
	min, max = get_drawn_range
	for i in min..max
	  @win.setpos(i - min, 0)
	  if (i == @index) then
	    if (defined? @win.attrset) then
	      # for curses-color.patch [ruby-list:18163]
	      if (@cursor_attr) then
		@win.attrset(@cursor_attr)
	      else
		@win.standout
	      end
	    else
	      @win.standout
	    end
	  end
	  @win.addstr(@list[i])
	  @win.addstr(" " * (@win.maxx - @win.curx))
	  if (i == @index) then
	    if (defined? @win.attrset) then
	      # for curses-color.patch [ruby-list:18163]
	      if (@cursor_attr) then
		if (@plain_attr) then
		  @win.attrset(@plain_attr)
		else
		  @win.attrset(A_NORMAL)
		end
	      else
		@win.standend
	      end
	    else
	      @win.standend
	    end
	  end
	end
	for i in (max - min + 1)..(@win.maxy - 1)
	  @win.setpos(i, 0)
	  @win.addstr(" " * @win.maxx)
	end
      end
    end
    private :draw_list

    def refresh
      draw_list
      @win.refresh
    end

    def redraw
      @win.clear
      refresh
    end

    def next_item
      @m.synchronize do
	if (@list.empty?) then
	  @index = 0
	else
	  @index = (@index + 1) % @list.length
	end
      end
    end

    def prev_item
      @m.synchronize do
	if (@list.empty?) then
	  @index = 0
	else
	  @index = (@index - 1) % @list.length
	end
      end
    end

    def next_page
      @m.synchronize do
	if (@list.empty?) then
	  @index = 0
	else
	  nitems = @win.maxy
	  npages = (@list.length - 1) / nitems + 1
	  @index = [
	    (@index + nitems) % (nitems * npages),
	    @list.length - 1
	  ].min
	end
      end
    end

    def prev_page
      @m.synchronize do
	if (@list.empty?) then
	  @index = 0
	else
	  nitems = @win.maxy
	  npages = (@list.length - 1) / nitems + 1
	  @index = [
	    (@index - nitems) % (nitems * npages),
	    @list.length - 1
	  ].min
	end
      end
    end

    def enter
      @m.synchronize do
	if (@list.empty?) then
	  throw(:CURSES_MENU, nil)
	else
	  throw(:CURSES_MENU, @index)
	end
      end
    end

    def cancel
      throw(:CURSES_MENU, nil)
    end

    def get
      catch(:CURSES_MENU) do
	begin
	  loop do
	    refresh
	    hook_key(@win.getch, self)
	  end
	ensure
	  if (defined? @win.attrset) then
	    # for curses-color.patch [ruby-list:18163]
	    if (@plain_attr || @cursor_attr) then
	      @win.attrset(A_NORMAL)
	    end
	  end
	end
      end
    end

    # [NOTE]
    # Keymap module method (callback.rb)
    # -
    # 1. set_keymap(key, callback)
    # 2. set_keymap(key) { |menu| ... }
    # -
    # callback ϼΥ᥽åɤʤФʤʤ
    #   callback.call(menu)
    #

    NEXT_ITEM = proc{ |menu| menu.next_item }
    PREV_ITEM = proc{ |menu| menu.prev_item }
    NEXT_PAGE = proc{ |menu| menu.next_page }
    PREV_PAGE = proc{ |menu| menu.prev_page }
    ENTER     = proc{ |menu| menu.enter }
    CANCEL    = proc{ |menu| menu.cancel }
    REDRAW    = proc{ |menu| menu.redraw }

    if (defined? Key) then
      # for curses.patch [ruby-list:16319]
      KEYMAP_TABLE = [
	# <key list>                     <callback>
	[ [ KEY_DOWN, ?N, ?n, ?J, ?j ], NEXT_ITEM ],
	[ [ KEY_UP, ?P, ?p, ?K, ?k ],   PREV_ITEM ],
	[ [ KEY_NPAGE, ?\C-f ],         NEXT_PAGE ],
	[ [ KEY_PPAGE, ?\C-b ],         PREV_PAGE ],
	[ [ ?\n ],                      ENTER ],
	[ [ ?\e, ?Q, ?q, ?X, ?x, ?/ ],  CANCEL ],
	[ [ ?\C-l ],                    REDRAW ]
      ]
    else
      KEYMAP_TABLE = [
	# <key list>              <callback>
	[ [ ?J, ?j, ?N, ?n ],     NEXT_ITEM ],
	[ [ ?K, ?k, ?P, ?p ],     PREV_ITEM ],
	[ [ ?\C-f, ?L, ?l ],      NEXT_PAGE ],
	[ [ ?\C-b, ?H, ?h ],      PREV_PAGE ],
	[ [ ?\n ],                ENTER ],
	[ [ ?Q, ?q, ?X, ?x, ?/ ], CANCEL ],
	[ [ ?\C-l ],              REDRAW ]
      ]
    end

    def set_keymap_table
      for keys, callback in KEYMAP_TABLE
	for key in keys
	  set_keymap(key, callback)
	end
      end
    end
    private :set_keymap_table
  end
end
