# $Id: mp3play.rb,v 1.27 2002/02/02 15:00:21 toki Exp $
# --
# :
#   Ruby  mp3 եդ뤿Υ饤֥ꡣ
#   mpg123 Ȥޥɥ饤Υץ졼ƤӤƤ롣
#
# 饹⥸塼:
#   class MPG123
#   class MP3Player
#   class MP3Player::MP3Info
#   class MP3Player::MPG123Info
#
# ƥ饹⥸塼δط:
#   [MPG123]
#     (none)
#   [MP3Player]
#     <include>
#       Callback (callback.rb)
#   [MP3Player::MP3Info]
#     (none)
#   [MP3Player::MPG123Info]
#     (none)
#
# ƥ饹⥸塼:
#   [MPG123]
#     CMD
#     SIGTERM
#     BREAK
#   [MP3Player]
#     BREAK
#   [MP3Player::MP3Info]
#     (none)
#   [MP3Player::MPG123Info]
#     (none)
#
# ƥ饹⥸塼Υ᥽å:
#   [MPG123]
#     <class method>
#       new(kcode=$KCODE)
#     <public method>
#       playing?
#       play(filename) { |line| ... }
#       pid { |pid| ... }
#       kill
#     <private method>
#       fork_mpg123(cmd)
#       set_fcntl_nonblock(io)
#       cut_line(buf) { |line| ... }
#   [MP3Player]
#     <class method>
#       new(mpg123_opts=nil, kcode=$KCODE)
#     <public method>
#       add_preplay_callback(callback)
#       add_preplay_callback { |player| ...callback... }
#       add_playing_callback(callback)
#       add_playing_callback { |player| ...callback... }
#       add_postplay_callback(callback)
#       add_postplay_callback { |player| ...callback... }
#       add_mp3_info_callback(callback)
#       add_mp3_info_callback { |player, mp3_info| ...callback... }
#       add_mpg123_info_callback(callback)
#       add_mpg123_info_callback { |player, mp3_info, mpg123_info| ...callback... }
#       remove_mp3_info_callback(callback)
#       remove_mpg123_info_callback(callback)
#       playing?
#       list
#       list=(play_list)
#       index
#       index=(index)
#       repeat
#       repeat=(repeat)
#       mp3_info
#       play(index=0)
#       stop
#       pause { ...waiting... }
#       next(step=1)
#       prev(step=1)
#       set_index(index)
#       set_index { |play_list, play_index| ... [return index] }
#       set_skip_frames(frames)
#       set_skip_frames { |mp3_info, mpg123_info| ... [return frames] }
#     <private method>
#       dispatch_mp3_info_callback(mp3_info)
#       dispatch_mpg123_info_callback(mp3_info, mpg123_info)
#       dispatch_preplay_callback
#       dispatch_playing_callback
#       dispatch_postplay_callback
#       init_mp3_info_index(playlist)
#       make_play_thread(mp3_info, mpg123_info, mpg123_opts=nil)
#       command
#   [MP3Player::MP3Info]
#     <class method>
#       new(filename)
#     <public method>
#       filename
#       index
#       index=(index)
#       read
#       title
#       artist
#       album
#       year
#       comment
#       genre
#       genre_num
#       version
#       version_num
#       layer
#       layer_num
#       error_protection
#       bitrate
#       sampling
#       mode
#       mode_num
#       copyright
#       original
#       frames
#       minutes
#       remaining_seconds
#       seconds
#       file_size
#   [MP3Player::MPG123Info]
#     <public method>
#       pid { |pid| ... }
#       clear
#       valid_frame?
#       valid_time?
#       valid_info?
#       line
#       line
#       frames
#       frames=(frames)
#       remaining_frames
#       remaining_frames=(frames)
#       seconds
#       seconds=(seconds)
#       remaining_seconds
#       remaining_seconds=(seconds)
#

require "fcntl"
require "thread"
require "rmp3/mp3info"
require "rmp3/callback"

class MPG123
  CMD = "mpg123"
  SIGTERM = 15
  BREAK = 0.1

  def initialize(kcode=$KCODE)
    @line_pattern = Regexp::new(%q/^.*?(\r|\n)/, false, kcode)
    @playing = false
    @kill_lock = Mutex::new
    @pid = nil
  end

  def playing?
    @playing
  end

  def fork_mpg123(cmd)
    pid = nil
    mpg123_in, mpg123_out = IO::pipe
    Thread::exclusive do
      pid = fork{
	mpg123_in.close
	STDERR.reopen(mpg123_out)
	exec(cmd)
      }
    end
    mpg123_out.close

    [ pid, mpg123_in ]
  end
  private :fork_mpg123

  def set_fcntl_nonblock(io)
    flags = io.fcntl(Fcntl::F_GETFL)
    flags |= Fcntl::O_NONBLOCK
    io.fcntl(Fcntl::F_SETFL, flags)
  end
  private :set_fcntl_nonblock

  # cut_line(buf) { |line| ... }
  def cut_line(buf)
    while (buf.sub!(@line_pattern, ''))
      yield($&)
    end
  end
  private :cut_line

  # play(filename) { |line| ... }
  def play(filename, mpg123_opts=nil)
    if (@playing) then
      raise "already playing."
    end
    @playing = true
    @kill_lock.lock
    begin
      begin
	siz = 1024
	buf = ""
	cmd = CMD
	cmd += " #{mpg123_opts}" if mpg123_opts
	cmd += " #{filename}"
	@pid, mpg123_in = fork_mpg123(cmd)
      ensure
	@kill_lock.unlock
      end

      begin
	set_fcntl_nonblock(mpg123_in)
	begin
	  loop do
	    buf << mpg123_in.sysread(siz)
	    cut_line(buf) do |line|
	      yield(line)
	    end
	  end
	rescue Errno::EWOULDBLOCK
	  sleep(BREAK)
	  retry
	rescue EOFError
	  # դνλ
	end
      ensure
	pid = nil
	@kill_lock.synchronize do
	  begin
	    Process::kill(SIGTERM, @pid)
	  rescue Errno::ESRCH
	    # ץ̵ä鲿⤷ʤ
	  end
	  pid = @pid
	  @pid = nil
	end
	Process::waitpid(pid, 0)

	cut_line(buf) do |line|
	  yield(line)
	end
	unless (buf.empty?) then
	  yield(buf + "\n")
	end
      end

      status = $? >> 8
      if (status != 0) then
	raise "mpg123 error(#{status}): #{cmd}"
      end
      status
    ensure
      @playing = false
    end
  end

  # pid { |pid| ... }
  def pid
    @kill_lock.synchronize do
      yield(@pid)
    end
  end

  def kill
    self.pid do |pid|
      if (pid) then
	Process::kill(SIGTERM, @pid)
      end
    end
  end
end

class MP3Player
  class MP3Info
    def initialize(filename)
      @filename = filename
      @tag = nil
      @mpeg = nil
      @read = false
    end

    attr_reader :filename
    attr_accessor :index

    def read
      @tag = MP3Tag::load(@filename) unless @tag
      @mpeg = MPEGAudioStreamFile::new(filename) unless @mpeg
      self
    end

    # get{ ... }
    def get
      read unless @read
      yield
    end
    private :get

    def title;             get{ @tag && @tag.title };      end
    def artist;            get{ @tag && @tag.artist };     end
    def album;             get{ @tag && @tag.album };      end
    def year;              get{ @tag && @tag.year };       end
    def comment;           get{ @tag && @tag.comment };    end
    def genre;             get{ @tag && @tag.genre };      end
    def genre_num;         get{ @tag && @tag.genre_num };  end
    def version;           get{ @mpeg.version };           end
    def version_num;       get{ @mpeg.version_num };       end
    def layer;             get{ @mpeg.layer };             end
    def layer_num;         get{ @mpeg.layer_num };         end
    def error_protection;  get{ @mpeg.error_protection };  end
    def bitrate;           get{ @mpeg.bitrate };           end
    def sampling;          get{ @mpeg.sampling };          end
    def mode;              get{ @mpeg.mode };              end
    def mode_num;          get{ @mpeg.mode_num };          end
    def copyright;         get{ @mpeg.copyright };         end
    def original;          get{ @mpeg.original };          end
    def frames;            get{ @mpeg.frames };            end
    def minutes;           get{ @mpeg.minutes };           end
    def remaining_seconds; get{ @mpeg.remaining_seconds }; end
    def seconds;           get{ @mpeg.seconds };           end
    def file_size;         get{ @mpeg.file_size };         end

    def to_s
      get do
	if (@tag && @tag.title && ! @tag.title.empty?) then
	  @tag.title
	else
	  @filename
	end
      end
    end
  end

  class MPG123Info
    def initialize(mpg123)
      @mpg123 = mpg123
      clear
    end

    # pid { |pid| ... }
    def pid(&block)
      @mpg123.pid(&block)
    end

    def clear
      @line = nil
      @frames = nil
      @remaining_frames = nil
      @seconds = nil
      @remaining_seconds = nil
    end

    def valid_frame?
      @frames && @remaining_frames
    end

    def valid_time?
      @seconds && @remaining_seconds
    end

    def valid_info?
      @line && valid_frame? && valid_time?
    end

    attr_accessor :line
    attr_accessor :frames
    attr_accessor :remaining_frames
    attr_accessor :seconds
    attr_accessor :remaining_seconds
  end

  include Callback

  BREAK = 0.1

  def initialize(mpg123_opts=nil, kcode=$KCODE)
    @playing = false
    @play_list = []
    @play_index = 0
    @repeat = false

    @mpg123 = MPG123::new(kcode)
    @mpg123_opts = "--verbose"
    @mpg123_opts += " #{mpg123_opts}" if mpg123_opts

    @mp3_info = nil
    @mpg123_info = MPG123Info::new(@mpg123)
    @mpg123_temp_opts = nil

    @preplay_callback = []
    @playing_callback = []
    @postplay_callback = []

    @mp3_info_callback = []
    @mpg123_info_callback = []

    @cmd_lock = Mutex::new
    @cmd = nil

    @frame_pattern =
      Regexp::new(%q/([0-9]+) \[ *([0-9]+)\]/, false, kcode)
    @time_pattern =
      Regexp::new(%q/([0-9]+:)?([0-9]+\.[0-9]+) \[([0-9]+:)?([0-9]+\.[0-9]+)\]/,
		  false, kcode)
  end

  #
  # 1. add_..._callback(callback)
  # 2. add_..._callback { |player| ... }
  # -
  # callback ϼΥ᥽åɤʤФʤʤ
  #   callback.call(player)
  #
  def add_preplay_callback(callback=nil, &block)
    add_callback(@preplay_callback, callback, &block)
  end

  def add_playing_callback(callback=nil, &block)
    add_callback(@playing_callback, callback, &block)
  end

  def add_postplay_callback(callback=nil, &block)
    add_callback(@postplay_callback, callback, &block)
  end

  #
  # 1. add_mp3_info_callback(callback)
  # 2. add_mp3_info_callback { |player, mp3_info| ...callback... }
  # -
  # callback ϼΥ᥽åɤʤФʤʤ
  #   callback.call(player, mp3_info)
  #
  def add_mp3_info_callback(callback=nil, &block)
    add_callback(@mp3_info_callback, callback, &block)
  end

  #
  # 1. add_mpg123_info_callback(callback)
  # 2. add_mpg123_info_callback { |player, mp3_info, mpg123_info| ...callback... }
  # -
  # callback ϼΥ᥽åɤʤФʤʤ
  #   callback.call(player, mp3_info, mpg123_info)
  #
  def add_mpg123_info_callback(callback=nil, &block)
    add_callback(@mpg123_info_callback, callback, &block)
  end

  def remove_preplay_callback(callback)
    remove_callback(@preplay_callback, callback)
  end

  def remove_playing_callback(callback)
    remove_callback(@playing_callback, callback)
  end

  def remove_postplay_callback(callback)
    remove_callback(@postplay_callback, callback)
  end

  def remove_mp3_info_callback(callback)
    remove_callback(@mp3_info_callback, callback)
  end

  def remove_mpg123_info_callback(callback)
    remove_callback(@mpg123_info_callback, callback)
  end

  def dispatch_preplay_callback
    dispatch_callback(@preplay_callback, self)
  end
  private :dispatch_preplay_callback

  def dispatch_playing_callback
    dispatch_callback(@playing_callback, self)
  end
  private :dispatch_playing_callback

  def dispatch_postplay_callback
    dispatch_callback(@postplay_callback, self)
  end
  private :dispatch_postplay_callback

  def dispatch_mp3_info_callback(mp3_info)
    dispatch_callback(@mp3_info_callback, self, mp3_info)
  end
  private :dispatch_mp3_info_callback

  def dispatch_mpg123_info_callback(mp3_info, mpg123_info)
    dispatch_callback(@mpg123_info_callback, self, mp3_info, mpg123_info.clone)
  end
  private :dispatch_mpg123_info_callback

  def playing?
    @playing
  end

  def list
    @play_list
  end

  def list=(play_list)
    @play_list = play_list.collect{ |mp3|
      if (mp3.kind_of? MP3Info) then
	mp3
      else
	MP3Info::new(mp3)
      end
    }
  end

  def index
    @play_index
  end

  def index=(index)
    if (playing?) then
      set_index(index)
    else
      @play_index = index
    end
  end

  attr_accessor :repeat

  def mp3_info
    if (playing?) then
      @mp3_info
    else
      @play_list[@play_index]
    end
  end

  def make_play_thread(mp3_info, mpg123_info, mpg123_opts=nil)
    options = @mpg123_opts
    options += " #{mpg123_opts}" if mpg123_opts

    line = nil
    frame = nil
    time = nil

    Thread::new{
      @mpg123.play(mp3_info.filename, options) do |line|
	mpg123_info.line = line
	if (line[-1] == ?\r) then
	  frame, time = line.split(",", 2)
	  if (frame =~ @frame_pattern) then
	    mpg123_info.frames = $1.to_i
	    mpg123_info.remaining_frames = $2.to_i
	  end
	  if (time =~ @time_pattern) then
	    $1.chop!
	    mpg123_info.seconds = $1.to_f * 60 + $2.to_f
	    $3.chop!
	    mpg123_info.remaining_seconds = $3.to_f * 60 + $4.to_f
	  end
	end
	dispatch_mpg123_info_callback(mp3_info, mpg123_info)
      end
    }
  end
  private :make_play_thread

  def init_mp3_info_index(playlist)
    playlist.each_with_index do |mp3_info, index|
      mp3_info.index = index
    end
  end
  private :init_mp3_info_index

  def play
    @cmd_lock.lock
    if (@playing) then
      raise "playing locked."
    end
    @playing = true
    index = @play_index
    begin
      catch(:MP3PLAYER_STOP) do
	loop do
	  dispatch_preplay_callback
	  init_mp3_info_index(@play_list)
	  begin
	    while (@play_index < @play_list.length)
	      @mp3_info = @play_list[@play_index]
	      @mpg123_info.clear

	      @mpg123_info.line = "\n"
	      dispatch_mp3_info_callback(@mp3_info)
	      dispatch_mpg123_info_callback(@mp3_info, @mpg123_info)

	      thread = make_play_thread(@mp3_info, @mpg123_info, @mpg123_temp_opts)
	      @mpg123_temp_opts = nil
	      sleep(BREAK) until @mpg123.playing?

	      begin
		catch(:MP3PLAYER_INTERRUPT) do
		  while (@mpg123.playing?)
		    begin
		      @cmd_lock.unlock
		      dispatch_playing_callback
		      sleep(BREAK)
		    ensure
		      @cmd_lock.lock
		    end

		    if (@cmd) then
		      cmd = @cmd
		      @cmd = nil
		      case (cmd)
		      when :PLAY
			throw :MP3PLAYER_INTERRUPT
		      when :STOP
			throw :MP3PLAYER_STOP
		      else
			raise "unknown command type: #{cmd}"
		      end
		    end
		  end
		  @play_index += 1
		end
	      ensure
		@mpg123.kill if @mpg123.playing?
		thread.join
	      end
	    end
	  ensure
	    dispatch_postplay_callback
	  end

	  if (@repeat) then
	    @play_index = 0
	  else
	    break
	  end
	end
      end
    ensure
      @mp3_info = nil
      @mpg123_info.clear
      @mpg123_temp_opts = nil

      @play_index = index
      @playing = false
      @cmd_lock.unlock
    end
  end

  def command
    catch (:CMD_EXIT) do
      while (playing?)
	@cmd_lock.synchronize do
	  unless (@cmd) then
	    throw(:CMD_EXIT, yield)
	  end
	  sleep(BREAK)
	end
      end
    end
  end
  private :command

  def stop
    command do
      @cmd = :STOP
      true
    end
  end

  # pause { ...waiting... }
  def pause
    command do
      if (@mpg123_info.valid_frame?) then
	@mpg123.kill if @mpg123.playing?
	yield
	@mpg123_info.line = "\n"
	dispatch_mpg123_info_callback(@mp3_info, @mpg123_info)
	@mpg123_temp_opts = "--skip #{@mpg123_info.frames}"
	@cmd = :PLAY
	true
      end
    end
  end

  def next(step=1)
    command do
      @cmd = :PLAY
      @play_index += step
    end
  end

  def prev(step=1)
    command do
      @cmd = :PLAY
      @play_index -= step
      @play_index = 0 if (@play_index < 0)
      @play_index
    end
  end

  # 1. set_index(index)
  # 2. set_index { |play_list, play_index| ... [return index] }
  def set_index(index=nil)
    command do
      index = yield(@play_list, @play_index) unless index
      if (index) then
	@cmd = :PLAY
	@play_index = index
      end
    end
  end

  # 1. set_skip_frames(frames)
  # 2. set_skip_frames { |mp3_info, mpg123_info| ... [return frames] }
  def set_skip_frames(frames=nil)
    command do
      if (@mpg123_info.valid_info?) then
	frames = yield(@mp3_info, @mpg123_info) unless frames
	if (frames) then
	  @cmd = :PLAY
	  @mpg123_temp_opts = "--skip #{frames}"
	  frames
	end
      end
    end
  end
end
