# $Id: property.rb,v 1.3 2000/09/24 09:55:31 toki Exp $
# --
# :
#  °ǡ١
#
# 饹⥸塼:
#   class Property
#   class Property::Group
#
# ƥ饹⥸塼δط:
#   [Property]
#     <include>
#       Enumerable
#   [Property::Group]
#     <include>
#       Enumerable
#
# ƥ饹⥸塼:
#   [Property]
#     (none)
#   [Property::Group]
#     (none)
#
# ƥ饹⥸塼Υ᥽å:
#   [Property]
#     <class method>
#       new
#     <public method>
#       add_group(name, parent_name=nil)
#       get_group(name)
#       get_property(name, attr)
#       set_property(name, attr, value)
#       self[name]
#       self[name, attr]
#       self[name, attr] = value
#       roots
#       each_root { |root| ... }
#       each_group { |group| ... }
#       each { |group, attr, value| ... }
#   [Property::Group]
#     <class method>
#       new(name, parent=nil)
#     <public method>
#       name
#       parent
#       children
#       self[attr]
#       self[attr] = value
#       each { |attr, value| ... }
#       each_children { |child| ... }
#

class Property
  class Group
    include Enumerable

    def initialize(name, parent=nil)
      @attrs = {}
      @name = name
      @parent = parent
      @children = []
    end

    attr :name
    attr :parent
    attr :children

    def [](attr)
      if (value = @attrs[attr]) then
	value
      elsif (@parent) then
	@parent[attr]
      end
    end

    def []=(attr, value)
      @attrs[attr] = value
    end

    # each { |attr, value| ... }
    def each
      for attr, value in @attrs
	yield(attr, value)
      end
    end

    # each_children { |child| ... }
    def each_children(&block)
      for child in @children
	yield(child)
	child.each_children(&block)
      end
    end
  end

  include Enumerable

  def initialize
    @groups = {}
    @roots = []
  end

  def add_group(name, parent_name=nil)
    if (@groups.has_key? name) then
      raise "already defined group: #{name}"
    end
    if (parent_name) then
      unless (@groups.has_key? parent_name) then
	raise "not defined parent group: #{parent}"
      end
      parent = @groups[parent_name]
      group = Group::new(name, parent)
      parent.children << group
    else
      group = Group::new(name)
      @roots << group
    end
    @groups[name] = group
    group
  end

  def get_group(name)
    @groups[name]
  end

  def get_property(name, attr)
    unless (@groups.has_key? name) then
      raise "not defined group: #{name}"
    end
    @groups[name][attr]
  end

  def set_property(name, attr, value)
    unless (@groups.has_key? name) then
      raise "not defined group: #{name}"
    end
    @groups[name][attr] = value
  end

  # 1. self[name]: get_group(name)
  # 2. self[name, attr]: get_property(name, attr)
  def [](*args)
    case (args.length)
    when 1
      get_group(*args)
    when 2
      get_property(*args)
    else
      raise "wrong arguments"
    end
  end

  def []=(name, attr, value)
    set_property(name, attr, value)
  end

  attr :roots

  # each_root { |root| ... }
  def each_root
    for root in @roots
      yield(root)
    end
  end

  # each_group { |group| ... }
  def each_group(&block)
    for root_group in @roots
      yield(root)
      root.each_children(&block)
    end
  end

  # each { |group, attr, value| ... }
  def each
    each_group do |group|
      for attr, value in group
	yield(group, attr, value)
      end
    end
  end
end
